import neovim

from ribosome import command, NvimStatePlugin

from amino.lazy import lazy
from amino import Left, L, _, Map
from ribosome.logging import Logging
from ribosome.request import function, msg_function, msg_command
from ribosome.machine import message, may_handle, handle, Machine
from ribosome.machine.state import RootMachine, RunScratchMachine
from ribosome.machine.transition import Fatal
from ribosome.machine.scratch import ScratchMachine, Mapping
from ribosome.nvim import NvimFacade, ScratchBuffer


Msg = message('Msg', 'text')
Err = message('Err')
Scratch = message('Scratch')
ScratchTest = message('ScratchTest')
ScratchCheck = message('ScratchCheck')


class Mach(RootMachine):

    @may_handle(Msg)
    def mess(self, data, msg):
        self.log.info(msg.text)

    @handle(Err)
    def err(self, data, msg):
        return Left(Fatal(TestPlugin.test_error))

    @may_handle(Scratch)
    def run_scratch(self, data, msg):
        ctor = L(ScratchM)(self.vim, _, _)
        return RunScratchMachine(ctor, False)

    @may_handle(ScratchCheck)
    def check_scratch(self, data, msg):
        self.log.info(self.sub.length)


class ScratchM(ScratchMachine):

    @property
    def prefix(self):
        return 'Test'

    @property
    def mappings(self):
        return Map()

    def __init__(self, vim: NvimFacade, scratch: ScratchBuffer,
                 parent: Machine) -> None:
        super().__init__(vim, scratch, parent=parent, title='scratch')

    @may_handle(ScratchTest)
    def test(self, data, msg):
        self.log.info(TestPlugin.test_scratch)


class TestPlugin(NvimStatePlugin, Logging):
    test_go = 'TestPlugin cmd test message'
    test_fun = 'TestPlugin fun test message'
    test_value = 'test value {}'
    test_error = 'test error'
    test_scratch = 'test scratch'

    def start(self):
        self.state.start()

    @lazy
    def state(self):
        return Mach(self.vim.proxy, title='spec')

    @neovim.function('Value', sync=True)
    def value(self, args):
        return self.test_value.format(args[0])

    @command(sync=True)
    def go(self):
        self.start()
        self.log.info(self.test_go)

    @function()
    def fun(self, value):
        return self.test_fun.format(value)

    @msg_function(Msg)
    def msg_fun(self):
        pass

    @msg_command(Err)
    def err(self):
        pass

    @msg_command(Scratch)
    def scratch(self):
        pass

    @msg_command(ScratchTest)
    def scratch_test(self):
        pass

    @msg_function(Mapping)
    def test_mapping(self):
        pass

    @msg_command(ScratchCheck)
    def check_scratch(self):
        pass

__all__ = ('TestPlugin',)



from ribosome.record import map_field, bool_field, Record, maybe_field

from amino import Empty, Just, Maybe, __
from ribosome.nvim.components import HasWindow
from ribosome.nvim import NvimIO, Tab


class ScratchBuilder(Record):
    params = map_field()
    use_tab = bool_field()
    vertical = bool_field()
    size = maybe_field(int)

    @property
    def tab(self):
        return self.set(use_tab=True)

    @property
    def build(self):
        return (
            NvimIO(self._setup_window)
            .map2(self._setup_buffer)
            .map3(self._create)
        )

    def _create_window(self, vim):
        return vim.vnew() if self.vertical else vim.new()

    def _setup_window(self, vim):
        if self.use_tab:
            tab = vim.tabnew()
            return Just(tab), tab.window
        else:
            return Empty(), self._create_window(vim)

    def _setup_buffer(self, tab, win):
        win.options.set_b('wrap', False)
        buffer = win.buffer
        buffer.options.set_s('buftype', 'nofile')
        buffer.options.set_s('bufhidden', 'wipe')
        buffer.options.set_b('buflisted', False)
        buffer.options.set_b('swapfile', False)
        self.size / 'resize {}'.format % win.cmd_sync
        return (tab, win, buffer)

    def _create(self, tab, win, buffer):
        return ScratchBuffer(win.vim, tab, win, buffer)


class ScratchBuffer(HasWindow):

    def __init__(self, vim, tab: Maybe[Tab], win, buffer) -> None:
        super().__init__(vim, buffer.target, buffer.prefix)
        self._tab = tab
        self._win = win
        self._buffer = buffer

    @property
    def _internal_window(self):
        return self._win.target

    @property
    def _internal_buffer(self):
        return self._buffer.target

    def set_content(self, text):
        self._buffer.set_modifiable(True)
        self._buffer.set_content(text)
        self._buffer.set_modifiable(False)

    def close(self):
        self._tab / __.close() | self._win.close()

__all__ = ('ScratchBuilder', 'ScratchBuffer')



from typing import Callable
from typing import Optional  # NOQA
import abc
import threading
import asyncio
from concurrent import futures
import importlib
from contextlib import contextmanager

from ribosome.logging import Logging
from ribosome.data import Data
from ribosome.nvim import HasNvim, ScratchBuilder
from ribosome import NvimFacade
from ribosome.machine.message_base import (message, Message, Nop, Done, Quit,
                                           PlugCommand, Stop)
from ribosome.machine.base import ModularMachine, Machine, Transitions
from ribosome.machine.transition import (TransitionResult, Coroutine,
                                         TransitionFailed, may_handle, handle)

from amino import Maybe, List, Map, may, Try, _, L, Empty, __

Callback = message('Callback', 'func')
IO = message('IO', 'perform')
Info = message('Info', 'message')
Envelope = message('Envelope', 'message', 'to')
RunMachine = message('RunMachine', 'machine')
KillMachine = message('KillMachine', 'uuid')
RunScratchMachine = message('RunScratchMachine', 'machine', 'tab',
                            opt_fields=(('size', Empty()),))
Init = message('Init')


class AsyncIOThread(threading.Thread, Logging, metaclass=abc.ABCMeta):

    def __init__(self) -> None:
        threading.Thread.__init__(self)
        self.done = None  # type: Optional[futures.Future]
        self.running = futures.Future()  # type: futures.Future

    def run(self):
        self._loop = asyncio.new_event_loop()  # type: ignore
        self._messages = asyncio.PriorityQueue(loop=self._loop)
        asyncio.set_event_loop(self._loop)
        self.done = futures.Future()
        self.running.set_result(True)
        try:
            self._loop.run_until_complete(self._main(self.init))
        except Exception:
            self.log.exception('while running state machine')
        self.running = futures.Future()

    def stop(self):
        if self.done is not None:
            self._stop()
            try:
                self.done.result(5)
            except Exception as e:
                self.log.error(e)
            finally:
                try:
                    self._loop.close()
                except Exception as e:
                    self.log.error(e)

    def _stop(self):
        self._done()

    def _done(self):
        self.done.set_result(True)

    @abc.abstractmethod
    async def _main(self, initial):
        ...

    @property
    def init(self):
        pass


class StateMachine(AsyncIOThread, ModularMachine):

    def __init__(self, sub: List[Machine]=List(), parent=None, title=None
                 ) -> None:
        AsyncIOThread.__init__(self)
        self.data = None
        self._messages = None
        self.sub = sub
        ModularMachine.__init__(self, parent, title=None)

    @property
    def init(self) -> Data:
        return self._data_type()

    def wait_for_running(self):
        self.running.result(3)

    def start_wait(self):
        self.start()
        self.wait_for_running()

    def _stop(self):
        self.send(Stop())

    def _publish(self, msg):
        return self._messages.put(msg)

    def send(self, msg: Message, prio=0.5):
        self.log.debug('send {}'.format(msg))
        if self._messages is not None:
            return asyncio.run_coroutine_threadsafe(  # type: ignore
                self._messages.put(msg), self._loop)

    def send_sync(self, msg: Message):
        self.send(msg)
        return self.await_state()

    def await_state(self):
        asyncio.run_coroutine_threadsafe(self.join_messages(),
                                         self._loop).result(2)
        return self.data

    def eval_expr(self, expr: str, pre: Callable=lambda a, b: (a, b)):
        sub = self.sub
        sub_map = Map(sub / (lambda a: (a.title, a)))
        data, plugins = pre(self.data, sub_map)
        return Try(eval, expr, None, dict(data=data, plugins=plugins))

    def _send(self, data, msg: Message):
        return (
            Maybe.from_call(
                self.loop_process, data, msg, exc=TransitionFailed) |
            TransitionResult.empty(data)
        )

    @may_handle(Nop)
    def _nop(self, data: Data, msg):
        pass

    @may_handle(Stop)
    def _stop_msg(self, data: Data, msg):
        return Quit(), Done().at(1)

    @may_handle(Done)
    def _done_msg(self, data: Data, msg):
        self._done()

    @may_handle(Callback)
    def message_callback(self, data: Data, msg):
        return msg.func(data)

    @may_handle(Coroutine)
    def _couroutine(self, data: Data, msg):
        return msg

    @may_handle(RunMachine)
    def _run_machine(self, data, msg):
        self.sub = self.sub.cat(msg.machine)
        return Envelope(Init(), msg.machine.uuid)

    @may_handle(KillMachine)
    def _kill_machine(self, data, msg):
        self.sub = self.sub.filter_not(_.uuid == msg.uuid)

    @handle(Envelope)
    def message_envelop(self, data, msg):
        return self.sub.find(_.uuid == msg.to) / __.loop_process(data,
                                                                 msg.message)

    def unhandled(self, data, msg):
        return self._fold_sub(data, msg)

    @may
    def _fold_sub(self, data, msg):
        ''' send **msg** to all sub-machines, passing the transformed
        data from each machine to the next and accumulating published
        messages.
        '''
        send = lambda z, s: z.accum(s.loop_process(z.data, msg))
        return self.sub.fold_left(TransitionResult.empty(data))(send)

    @contextmanager
    def transient(self):
        self.start()
        self.wait_for_running()
        self.send(Nop())
        yield self
        self.stop()

    async def _main(self, data):
        self.data = data
        while not self.done.done():
            self.data = await self._process_one_message(self.data)

    async def _process_one_message(self, data):
        msg = await self._messages.get()
        sent = self._send(data, msg)
        result = await sent.await_coro(self._dispatch_transition_result)
        for pub in result.pub:
            await self._publish(pub)
        self._messages.task_done()
        return result.data

    async def join_messages(self):
        await self._messages.join()


class PluginStateMachine(StateMachine):

    def __init__(self, plugins: List[str], title=None) -> None:
        StateMachine.__init__(self, title=title)
        self.sub = plugins.flat_map(self.start_plugin)

    @may
    def start_plugin(self, path: str):
        try:
            mod = importlib.import_module(path)
        except ImportError as e:
            msg = 'invalid {} plugin module "{}": {}'
            self.log.error(msg.format(self.title, path, e))
        else:
            if hasattr(mod, 'Plugin'):
                return getattr(mod, 'Plugin')(self.vim, self)

    def plugin(self, title):
        return self.sub.find(_.title == title)

    def plug_command(self, plug_name: str, cmd_name: str, args: list=[],
                     sync=False):
        sender = self.send_sync if sync else self.send
        plug = self.plugin(plug_name)
        cmd = plug.flat_map(lambda a: a.command(cmd_name, List(args)))
        plug.ap2(cmd, PlugCommand) % sender

    def plug_command_sync(self, *a, **kw):
        return self.plug_command(*a, sync=True, **kw)

    @may_handle(PlugCommand)
    def _plug_command(self, data, msg):
        self.log.debug(
            'sending command {} to plugin {}'.format(msg.msg, msg.plug.title))
        return msg.plug.process(data, msg.msg)


class RootMachine(PluginStateMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, plugins: List[str]=List(), title=None
                 ) -> None:
        HasNvim.__init__(self, vim)
        PluginStateMachine.__init__(self, plugins, title=None)

    @property
    def title(self):
        return 'ribosome'

    @handle(RunScratchMachine)
    def _scratch_machine(self, data, msg):
        return (
            ScratchBuilder(use_tab=msg.tab, size=msg.size)
            .build
            .unsafe_perform_io(self.vim) /
            L(msg.machine)(_, self) /
            RunMachine
        )


class SubMachine(ModularMachine):

    def new_state(self):
        pass


class SubTransitions(Transitions):

    def _state(self, data):
        return data.sub_state(self.name, self.machine.new_state)

    @property
    def state(self):
        return self._state(self.data)

    def _with_sub(self, data, state):
        return data.with_sub_state(self.name, state)

    def with_sub(self, state):
        return self._with_sub(self.data, state)

__all__ = ('Machine', 'Message', 'StateMachine', 'PluginStateMachine', 'Info')



from amino import Right
from amino.test import later

from ribosome.test import PluginIntegrationSpec

from integration._support.plugin import TestPlugin


class PluginSpec(PluginIntegrationSpec):

    @property
    def plugin_class(self):
        return Right(TestPlugin)

    def _last_output(self, content):
        later(lambda: self._log_out.last.should.contain(content), timeout=1)

    def startup(self):
        i = 99932
        fun_text = 'message function test'
        self.vim.cmd_sync('Go')
        val = TestPlugin.test_value.format(i)
        self.vim.call('Value', i).to_maybe.should.contain(val)
        self._last_output(TestPlugin.test_go)
        val2 = TestPlugin.test_fun.format(i)
        self.vim.call('Fun', i).to_maybe.should.contain(val2)
        self.vim.call('MsgFun', fun_text)
        self._last_output(fun_text)
        self.vim.cmd_sync('Err')
        self._last_output(TestPlugin.test_error)

    def scratch(self):
        self.vim.cmd_sync('Go')
        self.vim.cmd_sync('Scratch')
        later(lambda: self.vim.windows.length.should.equal(2))
        self.vim.cmd_sync('CheckScratch')
        self._last_output('1')
        self.vim.cmd_sync('ScratchTest')
        self._last_output(TestPlugin.test_scratch)
        self.vim.window.close()
        later(lambda: self.vim.windows.length.should.equal(1))
        self.vim.cmd_sync('CheckScratch')
        self._last_output('0')

__all__ = ('PluginSpec',)



from ribosome.machine.base import (Machine, ModularMachine, Transitions,
                                   RunTask, DataTask)
from ribosome.machine.state import (StateMachine, PluginStateMachine,
                                    RootMachine, Info)
from ribosome.machine.transition import (TransitionFailed, handle, may_handle,
                                         either_msg, either_handle, Error)
from ribosome.machine.message_base import (Nop, Message, message, json_message,
                                           Quit)

__all__ = ('Machine', 'StateMachine', 'PluginStateMachine', 'handle',
           'may_handle', 'TransitionFailed', 'RootMachine', 'ModularMachine',
           'Nop', 'either_msg', 'either_handle', 'Transitions', 'Message',
           'Error', 'RunTask', 'message', 'json_message', 'DataTask', 'Quit',
           'Info')



rplugin_template = '''
import neovim
import os
import logging
from pathlib import Path

from {plugin_module} import {plugin_class}

import amino

amino.development = True

import amino.logging

logfile = Path(os.environ['RIBOSOME_LOG_FILE'])
amino.logging.amino_file_logging(level=logging.DEBUG,
                               handler_level=logging.DEBUG,
                               logfile=logfile)


@neovim.plugin
class Plugin({plugin_class}):
    pass
'''

__all__ = ('rplugin_template',)
